/* ----
 * ---- file   : c64_tc.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the GNU GENERAL PUBLIC LICENSE (GPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#GPL or COPYING for further information.
 * ----
 * ---- info   : This is part of the "c64_tools" package.
 * ----
 * ---- changed: 12Sep2013, 15Sep2013, 16Sep2013, 17Sep2013, 18Sep2013, 19Sep2013, 20Sep2013
 * ----
 * ----
 */

#define NUM_BENCHMARK_ITERATIONS (100000u)
//#define NUM_BENCHMARK_ITERATIONS (10u)


#define NUM_FASTCALL_BENCHMARK_ITERATIONS (1000000u)


#define NUM_SRAM_ITERATIONS (10000000u)
#define SRAM_REG(a) (*(volatile sU32*)(fshm.virt_addr + (a)))


//#define SHM_SIZE (4u * 1024u * 1024u)
#define SHM_SIZE (4u * 1024u * 2)
//#define SHM_SIZE (1u * 1024u * 1024u)


// Testcase enumeration:
#define TC_NONE                        (0u)
#define TC_TESTMSG                     (1u)  // Send 4 byte test message to DSP, bypassing client handlers
#define TC_RPC_ADD                     (2u)  // Perform single remote procedure call (add two signed 32bit ints)
#define TC_RPC_ADD_BENCHMARK           (3u)  // RPC stresstest, repeatedly calls 'add' cmd of 'demo_calc' component
#define TC_IOCTL_TEST                  (4u)  // Call C64_IOCTL_TEST (debugging)
#define TC_SHM_ACCESS                  (5u)  // Tests shared memory access (read/write)
#define TC_RPC_ADD_FASTCALL_BENCHMARK  (6u)
#define TC_SRAM_SPEED                  (7u)

#define TC_NUM                         (8u)


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "../include/inc_libc64.h"

#include <c64p_simple_dm3730/pjt/components.h>


/*--------------------------------------------------------------------------- module vars */
#define TC_STR(a) #a
static const char *tc_names[TC_NUM] = {
   TC_STR(TC_NONE),
   TC_STR(TC_TESTMSG),
   TC_STR(TC_RPC_ADD),
   TC_STR(TC_RPC_ADD_BENCHMARK),
   TC_STR(TC_IOCTL_TEST),
   TC_STR(TC_SHM_ACCESS),
   TC_STR(TC_FASTCALL_RPC_ADD),
   TC_STR(TC_SRAM_SPEED)
};
#undef TC_STR


static dsp_mem_region_t shm;
static dsp_mem_region_t fshm;


/*--------------------------------------------------------------------------- usage() */
static void usage(void) {
   sUI tcIdx;
   
   printf("Usage:\n");
   printf("  c64_tc [testcase_nr]\n\n");
   printf("    testcase_nr can be one of:\n");
   
   for(tcIdx=1; tcIdx < TC_NUM; tcIdx++)
   {
      printf("\t\t%02u: %s\n",
             tcIdx,
             tc_names[tcIdx]
             );
   }
   
   puts("");
}


/*--------------------------------------------------------------------------- loc_tc_testmsg() */
static void loc_tc_testmsg(void) {
   /* Send a test message bypassing the client handler in the c64 kernel module 
    *
    *  This is just to see whether the GPP<>DSP messaging in general works OK. 
    *
    *   (note) the kernel module should have been loaded with "debug=20", i.e.
    *           $ /sbin/insmod c64.ko debug=20
    */
   dsp_send_forced_message(0x37703370u);
}


/*--------------------------------------------------------------------------- loc_tc_rpc_add_test() */
static void loc_tc_rpc_add_test(void) {
   /* Test single DSP remote procedure call */
   c64_msg_t msg;
   c64_msg_t reply;
   int ret;
   
   if(LOG_DEBUG_LEVEL(1))
   {
      log_printf(LOG_DEBUG "sizeof(c64_msg_t) == %u\n", sizeof(c64_msg_t)); // must be 12 bytes
   }
   
   msg.req.fields.dsp_component_id = COMPONENT_DEMO_CALC; // selects "demo_calc" component on DSP-side
   msg.req.fields.dsp_cmd_id       = DEMO_CALC_CMD_ADD;   // selects the component's 'ADD' command
   
   msg.data[0].arg = 42 + (rand()&31);             // argument #1
   msg.data[1].arg = 22 + (rand()&31);             // argument #2
   
   ret = dsp_rpc(&msg, &reply);
   
   if(LOG_DEBUG_LEVEL(10))
   {
      dsp_debug_internal_print();
   }
   
   if(0 == ret)
   {
      log_printf(LOG_INFO "loc_rpc() OK: result=%d (expected %d)\n", 
                 *(sS32*) &reply.data[0].ret,   /* re-interpet unsigned int as signed */
                 (msg.data[0].arg + msg.data[1].arg)
                 );
      
      if(LOG_DEBUG_LEVEL(10))
      {
         dsp_debug_internal_print();
      }
   }
   else
   {
      log_printf(LOG_ERROR "loc_rpc() failed (ret=%d)\n", ret);
   }
}


/*--------------------------------------------------------------------------- loc_tc_rpc_add_benchmark() */
static void loc_tc_rpc_add_benchmark(void) {

   log_printf(LOG_INFO "starting benchmark\n");
   {
      c64_msg_t msg;
      c64_msg_t reply;
      sUI       tStart = osal_milliseconds_get();
      sUI       tEnd;
      sUI       tDelta;
      sUI       i;
      int       r;

      msg.req.fields.dsp_component_id = COMPONENT_DEMO_CALC; // selects "demo_calc" component on DSP-side
      msg.req.fields.dsp_cmd_id       = DEMO_CALC_CMD_ADD;   // selects the component's 'ADD' command
      
      msg.data[0].arg = 42;                           // argument #1
      msg.data[1].arg = 0;                            // argument #2

      for(i=0; i<NUM_BENCHMARK_ITERATIONS; i++)
      {
         r = dsp_rpc(&msg, &reply);

         if(0 == r)
         {
            if(reply.data[0].ret != (msg.data[0].arg + msg.data[1].arg))
            {
               log_printf(LOG_ERROR "iteration=%u: unexpected RPC result: have=0x%08x, expected=0x%08x\n",
                          i,
                          reply.data[0].ret,
                          (msg.data[0].arg + msg.data[1].arg)
                          );
               
               dsp_debug_internal_print();
               break;
            }
         }
         else
         {
            log_printf(LOG_ERROR "iteration=%u: loc_rpc() failed (r=%d).\n", r);

            dsp_debug_internal_print();
            break;
         }


         /* Next iteration */
         msg.data[0].arg++;
         msg.data[1].arg++;
      }

      if(NUM_BENCHMARK_ITERATIONS == i)
      {

         tEnd = osal_milliseconds_get();
         
         tDelta = tEnd - tStart;
         
         log_printf(LOG_INFO "benchmark finished. => %u iterations in %u millisecs.\n",
                    NUM_BENCHMARK_ITERATIONS,
                    tDelta
                    );
      }
      else
      {
         log_printf(LOG_ERROR "(benchmark aborted due to error)\n");
      }
   }
}


/*--------------------------------------------------------------------------- loc_tc_shm_access_test() */
static void loc_tc_shm_access_test(void) {
   sUI i;
   sU32 *d;
   int ret;
   sU32 chkGPP = 0;

   d = (sU32*) shm.virt_addr;

   dsp_debug_usr_print(0);

   log_printf(LOG_INFO "loc_tc_shm_access_test: virtAddr=0x%p, trying to write..\n", d);

   for(i=0; i<(shm.size >> 2); i++)
   {
      d[i] = i;
      chkGPP += i;
   }

   log_printf(LOG_INFO "loc_tc_shm_access_test: OK, wrote %d dwords (%u kBytes).\n", i, (i >> 8));

   /* Writeback GPP caches so DSP can see the updated data */
   ret = dsp_cache_wb(shm.phys_addr,
                      shm.size
                      );

   if(0 == ret)
   {
      /* Now verify that DSP sees the same data by calling the DSP-side demo_checksum::add() method */
      sU32 chkDSP = 0xCCccCCccu;

      ret = dsp_rpc_2a_1r(COMPONENT_DEMO_CHECKSUM,
                          DEMO_CHECKSUM_ADD,
                          shm.phys_addr,
                          shm.size,
                          &chkDSP
                          );

      if(LOG_DEBUG_LEVEL(10))
      {
         dsp_debug_internal_print();
      }

      if(chkDSP == chkGPP)
      {
         log_printf(LOG_INFO "loc_tc_shm_access_test: OK, DSP+GPP checksums match (chk=0x%08x)\n",
                    chkDSP
                    );
      }
      else
      {
         log_printf(LOG_ERROR "loc_tc_shm_access_test: failed, checksum mismatch (gpp=0x%08x dsp=0x%08x)\n",
                    chkGPP,
                    chkDSP
                    );

         dsp_debug_usr_print(0);
      }
   }
   else
   {
      log_printf(LOG_ERROR "loc_tc_shm_access_test: loc_cache_wb() failed (ret=%d)\n", ret);
   }
}


/*--------------------------------------------------------------------------- loc_tc_rpc_fastcall_add_benchmark() */
static void loc_tc_rpc_add_fastcall_benchmark(void) {

   log_printf(LOG_INFO "starting fastcall benchmark\n");
   {
      c64_msg_t msg;
      c64_msg_t reply;
      sUI       tStart = osal_milliseconds_get();
      sUI       tEnd;
      sUI       tDelta;
      sUI       i;
      int       r;

      msg.req.fields.dsp_component_id = COMPONENT_DEMO_CALC_FASTCALL; // selects "demo_calc_fastcall" component on DSP-side
      msg.req.fields.dsp_cmd_id       = DEMO_CALC_FASTCALL_INITIATE;  // selects the component's 'INITIATE' command

      /* Initiate fastcall sequence */
      /* (note) do not care for args */
      r = dsp_rpc_send(&msg);

      printf("xxx fastcall_benchmark: init rpc sent r=%d\n", r);

      if(0 == r)
      {
         sU32 arg1 = 42;
         sU32 arg2 = 0;

         for(i=0; i<NUM_FASTCALL_BENCHMARK_ITERATIONS; i++)
         {
            sU32 r1FC;
            sU32 r2FC;

            r = dsp_fastcall_rpc(DEMO_CALC_FASTCALL_FC_CMD_ADD,
                                 arg1, arg2,
                                 &r1FC, &r2FC
                                 );
            
            if(0 == r)
            {
               if(r1FC != (arg1 + arg2))
               {
                  log_printf(LOG_ERROR "iteration=%u: unexpected fastcall RPC result: have=0x%08x, expected=0x%08x\n",
                             i,
                             r1FC,
                             (arg1 + arg2)
                             );

                  dsp_fastcall_end();
                  
                  dsp_debug_internal_print();
                  break;
               }
            }
            else
            {
               log_printf(LOG_ERROR "iteration=%u: dsp_fastcall_rpc() failed (r=%d).\n", r);
               
               dsp_fastcall_end();

               dsp_debug_internal_print();
               break;
            }


            /* Next iteration */
            arg1++;
            arg2++;
         }

         if(NUM_FASTCALL_BENCHMARK_ITERATIONS == i)
         {
            dsp_fastcall_end();

            tEnd = osal_milliseconds_get();
         
            tDelta = tEnd - tStart;
         
            log_printf(LOG_INFO "fastcall benchmark finished. => %u iterations in %u millisecs.\n",
                       NUM_FASTCALL_BENCHMARK_ITERATIONS,
                       tDelta
                       );
         }
         else
         {
            log_printf(LOG_ERROR "(fastcall benchmark aborted due to error)\n");
         }

         /* Pick up reply from DSP (can be omitted) */
         dsp_rpc_recv(&reply);
      }
      else
      {
         /* Failed: dsp_rpc_send() */
      }
   }
}


/*--------------------------------------------------------------------------- loc_tc_sram_speed() */
static void loc_tc_sram_speed(void) {
   sUI i;
   sUI tStart;
   sUI tEnd;
   sU32 readMem[16];
   sU32 chksum;

   /* Write */
   log_printf(LOG_INFO "starting SRAM write benchmark..");

   tStart = osal_milliseconds_get();

   for(i=0; i<(NUM_SRAM_ITERATIONS >> 4u); i++)
   {
      SRAM_REG( 0) = 0x00000000u;
      SRAM_REG( 4) = 0x11111111u;
      SRAM_REG( 8) = 0x22222222u;
      SRAM_REG(12) = 0x33333333u;
                  
      SRAM_REG(16) = 0x44444444u;
      SRAM_REG(20) = 0x55555555u;
      SRAM_REG(24) = 0x66666666u;
      SRAM_REG(28) = 0x77777777u;
                  
      SRAM_REG(32) = 0x88888888u;
      SRAM_REG(36) = 0x99999999u;
      SRAM_REG(40) = 0xaaaaaaaau;
      SRAM_REG(44) = 0xbbbbbbbbu;
                  
      SRAM_REG(48) = 0xccccccccu;
      SRAM_REG(52) = 0xddddddddu;
      SRAM_REG(56) = 0xeeeeeeeeu;
      SRAM_REG(60) = 0xffffffffu;
   }

   tEnd = osal_milliseconds_get();

   log_printf("..DONE: %u 32bit write accesses in %u milliseconds\n",
              NUM_SRAM_ITERATIONS,
              (tEnd - tStart)
              );


   /* Read */
   log_printf(LOG_INFO "starting SRAM read benchmark..");

   tStart = osal_milliseconds_get();

   for(i=0; i<(NUM_SRAM_ITERATIONS >> 4u); i++)
   {
      readMem[ 0] = SRAM_REG( 0);
      readMem[ 1] = SRAM_REG( 4);
      readMem[ 2] = SRAM_REG( 8);
      readMem[ 3] = SRAM_REG(12);
                                      
      readMem[ 4] = SRAM_REG(16);
      readMem[ 5] = SRAM_REG(20);
      readMem[ 6] = SRAM_REG(24);
      readMem[ 7] = SRAM_REG(28);
                                
      readMem[ 8] = SRAM_REG(32);
      readMem[ 9] = SRAM_REG(36);
      readMem[10] = SRAM_REG(40);
      readMem[11] = SRAM_REG(44);
                                
      readMem[12] = SRAM_REG(48);
      readMem[13] = SRAM_REG(52);
      readMem[14] = SRAM_REG(56);
      readMem[15] = SRAM_REG(60);
   }

   tEnd = osal_milliseconds_get();

   /* Prevent compiler from optimization / discarding reads (maybe not necessary?) */
   chksum = 0;
   for(i=0; i<16; i++)
   {
      //printf("\nxxx readMem[%u] = 0x%08x\n", i, readMem[i]);
      chksum += readMem[i];
   }

   log_printf("..DONE: %u 32bit read accesses in %u milliseconds (chksum=0x%08x)\n",
              NUM_SRAM_ITERATIONS,
              (tEnd - tStart),
              chksum
              );
}


/*--------------------------------------------------------------------------- main() */
int main(int argc, char**argv) {
   int ret;

   if(1 == argc)
   {
      usage();

      return 0;
   }

   /* Open client connection */
   ret = dsp_open();

   if(0 == ret)
   {
      sUI tcIdx = TC_NONE;

      /* Determine selected testcase */
      if(2 == argc)
      {
         sscanf(argv[1], "%u", &tcIdx);

         if(tcIdx >= TC_NUM)
         {
            log_printf(LOG_ERROR "invalid testcase idx %u\n", tcIdx);

            tcIdx = TC_NONE;
         }
      }

      log_printf(LOG_INFO "selected testcase %u (\"%s\")\n", tcIdx, tc_names[tcIdx]);

      if(LOG_DEBUG_LEVEL(10))
      {
         dsp_debug_internal_print();
      }

      ret = 0;

      /* Allocate contiguous memory block */
      if(0 == dsp_shm_init(SHM_SIZE))
      {
         /* Query shared memory attributes */
         shm = dsp_shm_get();

         /* Query fast shared memory attributes */
         fshm = dsp_fshm_get();


         /* Run testcase */
         switch(tcIdx)
         {
            default:
               break;
               
            case TC_TESTMSG:
            {
               loc_tc_testmsg();
            }
            break;
            
            case TC_RPC_ADD:
            {
               loc_tc_rpc_add_test();
            }
            break;
            
            
            case TC_RPC_ADD_BENCHMARK:
            {
               /* Benchmark DSP remote procedure call speed */
               
               loc_tc_rpc_add_benchmark();
               
            }
            break;
            
            case TC_IOCTL_TEST:
            {
               dsp_ioctl_test();
            }
            break;
            
            case TC_SHM_ACCESS:
            {
               loc_tc_shm_access_test();
            }
            break;

            case TC_RPC_ADD_FASTCALL_BENCHMARK:
            {
               loc_tc_rpc_add_fastcall_benchmark();
            }
            break;

            case TC_SRAM_SPEED:
            {
               loc_tc_sram_speed();
            }
            break;
            
         } // switch

         /* Free contiguous memory block */
         dsp_shm_exit();
         
      } /* if dsp_shm_init() */
         
      /* Disconnect client */
      dsp_close();

   } /* if dsp_open() */

   return ret;
}
